/* Plugin for Total Commander handling UltraCompressor II archives.
   Copyright © Jan Bobrowski 2021
   torinak.com/~jb/unuc2/

   This program is free software; you can redistribute it and
   modify it under the terms of the GNU Lesser General Public
   License version 3 as published by the Free Software Foundation.
*/

#include <windows.h>
#include <stdio.h>

// git clone https://github.com/ghisler/WCX-SDK.git
#include "WCX-SDK/src/wcxhead.h"

#include "libunuc2.h"
#include "list.h"

#define elemof(T) (sizeof T/sizeof*T)

struct uc2_wcx {
	uc2_handle uc2;
	struct list entries;
	struct list *current;
	FILE *file;
};

static int my_read(void *ctx, unsigned pos, void *ptr, unsigned len)
{
	struct uc2_wcx *wcx = ctx;
	if (fseek(wcx->file, pos, SEEK_SET) < 0)
		return -1;
	return fread(ptr, 1, len, wcx->file);
}

static void *my_alloc(void *ctx, unsigned size)
{
	return malloc(size);
}

static void my_free(void *ctx, void *ptr)
{
	free(ptr);
}

static struct uc2_io io = {
	.read = my_read,
	.alloc = my_alloc,
	.free = my_free
};

struct entry {
	struct list list;
	unsigned version;
	struct uc2_entry entry;
};

static int get_entry(struct entry *e, uc2_handle uc2);

static struct uc2_wcx *open_archive(FILE *f, int *result)
{
	int r = E_EOPEN;
	if (!f)
		goto error;
	struct uc2_wcx *wcx = malloc(sizeof *wcx);
	r = E_NO_MEMORY;
	if (!wcx)
		goto error_close;
	list_init(&wcx->entries);
	wcx->current = &wcx->entries;
	wcx->file = f;

	uc2_handle uc2 = uc2_open(&io, wcx);
	if (!uc2)
		goto error_free;
	wcx->uc2 = uc2;

	for (;;) {
		struct entry *e = malloc(sizeof *e);
		if (!e)
			break;
		r = get_entry(e, uc2);
		if (r <= 0) {
			free(e);
			if (r == UC2_End) {
				char label[12];
				uc2_finish_cdir(uc2, label);
			}
			break;
		}
		for (struct list *l = wcx->entries.next; l != &wcx->entries; l = l->next) {
			struct entry *e0 = list_item(l, struct entry, list);
			if (e->entry.name_len == e0->entry.name_len
			 && memcmp(e->entry.name, e0->entry.name, e->entry.name_len) == 0)
				e0->version++;
		}
		e->version = 0;
		list_append(&wcx->entries, &e->list);
	}

	if (r >= 0 || !list_is_empty(&wcx->entries))
		return wcx;

	r = E_BAD_ARCHIVE;
	uc2_close(uc2);
error_free:
	free(wcx);
error_close:
	fclose(f);
error:
	*result = r;
	return 0;
}

__declspec(dllexport) void *__stdcall OpenArchive(tOpenArchiveData *ad)
{
	FILE *f = fopen(ad->ArcName, "rb");
	return open_archive(f, &ad->OpenResult);
}

__declspec(dllexport) void *__stdcall OpenArchiveW(tOpenArchiveDataW* ad)
{
	FILE *f = _wfopen(ad->ArcName, L"rb");
	return open_archive(f, &ad->OpenResult);
}

static int get_entry(struct entry *e, uc2_handle uc2)
{
	int r = uc2_read_cdir(uc2, &e->entry);
	if (r <= 0)
		return r;
	while (r == UC2_TaggedEntry) {
		char tag[16];
		r = uc2_get_tag_header(uc2, tag);
		if (r < 0)
			return r;
		char data[r];
		r = uc2_get_tag_data(uc2, &e->entry, data);
		if (r < 0)
			return r;
	}
	return 1;
}

static struct entry *next_entry(struct uc2_wcx *wcx)
{
	struct list *l = wcx->current->next;
	wcx->current = l;
	if (l == &wcx->entries)
		return 0;
	return list_item(l, struct entry, list);
}

static char *prepare_name(struct uc2_wcx *wcx, char *buf, unsigned space, struct entry *e)
{
	char *p = buf + space;
	*--p = 0;

	if (e->version) {
		p -= _scprintf(";%u", e->version);
		_snprintf(p, INT_MAX, ";%u", e->version);
	}

	for (;;) {
		unsigned l = e->entry.name_len;
		if (p - buf < l)
			return p;
		p -= l;
		memcpy(p, e->entry.name, l);
		unsigned dirid = e->entry.dirid;
		if (!dirid || p - buf < 2)
			return p;

		*--p = '\\';

		for (struct list *l = wcx->entries.next; ; l = l->next) {
			if (l == &wcx->entries) {
				*--p = '?';
				return p;
			}
			e = list_item(l, struct entry, list);
			if (e->entry.id == dirid)
				break;
		}
	}
}

__declspec(dllexport) int __stdcall ReadHeader(void *handle, tHeaderData *hd)
{
	struct uc2_wcx *wcx = handle;
	struct entry *e = next_entry(wcx);
	if (!e)
		return E_END_ARCHIVE;

	char buf[elemof(hd->FileName)];
	char *name = prepare_name(wcx, buf, sizeof buf, e);
	memcpy(hd->FileName, name, buf + sizeof buf - name);

	hd->FileAttr = e->entry.attr;
	hd->UnpSize = e->entry.size;
	hd->PackSize = e->entry.csize;
	hd->FileTime = e->entry.dos_time;
	return 0;
}

__declspec(dllexport) int __stdcall ReadHeaderExW(void *handle, tHeaderDataExW *hd)
{
	struct uc2_wcx *wcx = handle;
	struct entry *e = next_entry(wcx);
	if (!e)
		return E_END_ARCHIVE;

	char buf[1024];
	char *name = prepare_name(wcx, buf, sizeof buf, e);

	MultiByteToWideChar(CP_UTF8, 0, name, -1, hd->FileName, elemof(hd->FileName));
	hd->FileAttr = e->entry.attr;
	hd->UnpSize = e->entry.size; hd->PackSizeHigh = 0;
	hd->PackSize = e->entry.csize; hd->UnpSizeHigh = 0;
	hd->FileTime = e->entry.dos_time;
	return 0;
}

static int write_file(void *file, const void *ptr, unsigned len)
{
	if (file) {
		if (fwrite(ptr, 1, len, file) < len)
			return -1;
	}
	return 0;
}

__declspec(dllexport) int __stdcall ProcessFile(void *handle, int op, char *DestPath, char *DestName)
{
	if (op == PK_SKIP)
		return 0;

	struct uc2_wcx *wcx = handle;
	FILE *f = 0;
	if (op == PK_EXTRACT) {
		f = fopen(DestName, "wb");
		if (!f)
			return E_ECREATE;
	}
	struct entry *e = list_item(wcx->current, struct entry, list);
	int r = uc2_extract(wcx->uc2, &e->entry, write_file, f);
	if (f)
		fclose(f);
	if (r < 0)
		return r == UC2_UserFault ? E_EWRITE : E_BAD_DATA;
	return 0;
}

__declspec(dllexport) int __stdcall CloseArchive(void *handle)
{
	struct uc2_wcx *wcx = handle;
	uc2_close(wcx->uc2);
	fclose(wcx->file);
	for (struct list *l = wcx->entries.next; l != &wcx->entries;) {
		struct entry *e = list_item(l, struct entry, list);
		l = l->next;
		free(e);
	}
	free(wcx);
	return 0;
}

__declspec(dllexport) void __stdcall SetChangeVolProc(void *handle, tChangeVolProc pChangeVolProc1) {}

__declspec(dllexport) void __stdcall SetProcessDataProc(void *handle, tProcessDataProc pProcessDataProc) {}
